import os
import sys
import logging
from logging import StreamHandler, Formatter, INFO, WARN, ERROR


def set_streamhandler(logger):
    logger.setLevel(INFO)
    _handler = StreamHandler(sys.stdout)
    _handler.setFormatter(
        Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    )
    logger.addHandler(_handler)


DISABLE_LOGGER_STREAMOUTPUT = "QM_DISABLE_STREAMOUTPUT"
logger = logging.getLogger("qm")
if os.environ.get(DISABLE_LOGGER_STREAMOUTPUT, None) is None:
    set_streamhandler(logger)

__all__ = ["INFO", "WARN", "ERROR", "logger"]



import numpy as np
from typing import Type, Collection

from qm._QmJobErrors import _handle_job_manager_error
from qm.logger import logger, INFO, WARN, ERROR
from qm.pb.errors_pb2 import (
    MissingJobError,
    InvalidJobExecutionStatusError,
    UnknownInputStreamError,
)
from qm.pb.general_messages_pb2 import (
    Message_LEVEL_ERROR,
    Message_LEVEL_INFO,
    Message_LEVEL_WARNING,
)
from qm.pb.job_manager_pb2 import InsertInputStreamRequest


def fix_object_data_type(obj):
    if isinstance(obj, (np.floating, np.integer, np.bool_)):
        return obj.item()
    else:
        return obj


_level_map = {
    Message_LEVEL_ERROR: ERROR,
    Message_LEVEL_WARNING: WARN,
    Message_LEVEL_INFO: INFO,
}


def _set_compiler_options(request, **kwargs):
    request.highLevelProgram.compilerOptions.optimizeMergeCodeExecution = (
        kwargs.get("optimize_merge_code_execution") is not False
    )
    request.highLevelProgram.compilerOptions.optimizeWriteReadCommands = (
        kwargs.get("optimize_read_write_commands") is True
    )
    request.highLevelProgram.compilerOptions.strict = (
        kwargs.get("strict", False) is True
    )

    # handle deprecated skip_optimizations (backwards compatible for future flags)
    skip_optimizations = kwargs.get("skip_optimizations", None)
    if skip_optimizations is None:
        pass
    elif type(skip_optimizations) is tuple:
        optimization_to_skip = []
        for opt in skip_optimizations:
            if type(opt) is str:
                request.highLevelProgram.compilerOptions.skipOptimizations.append(opt)
                optimization_to_skip.append(opt)
        logger.info("Skipping optimizations: " + ",".join(optimization_to_skip))
    else:
        logger.warn("skip_optimizations must be a tuple of strings")

    # handle deprecated extra_optimizations (backwards compatible for future flags)
    extra_optimizations = kwargs.get("extra_optimizations", None)
    if extra_optimizations is None:
        pass
    elif type(extra_optimizations) is tuple:
        optimization_to_add = []
        for opt in extra_optimizations:
            if type(opt) is str:
                request.highLevelProgram.compilerOptions.skipOptimizations.append(
                    "!" + opt
                )
                optimization_to_add.append(opt)
        logger.info("extra optimizations: " + ",".join(optimization_to_add))
    else:
        logger.warn("extra_optimizations must be a tuple of strings")

    try:
        flags_arg = kwargs.get("flags", [])
        flags = [opt for opt in flags_arg if type(opt) is str]
    except TypeError:
        flags = []
    for opt in flags:
        request.highLevelProgram.compilerOptions.flags.append(opt)
    logger.info("Flags: " + ",".join(flags))


def insert_input_stream_to_job(job, data, input_stream):
    if not job._qmm._caps.supports_input_stream:
        raise Exception("`insert_input_stream()` is not supported by the QOP version.")
    request = InsertInputStreamRequest()
    request.jobId = job._id
    request.streamName = "input_stream_" + input_stream
    if not isinstance(data, list):
        data = [data]
    if isinstance(data[0], int):
        request.intStreamData.data.extend(data)
        request.intStreamData.SetInParent()
    elif isinstance(data[0], float):
        request.fixedStreamData.data.extend(data)
        request.fixedStreamData.SetInParent()
    elif isinstance(data[0], bool):
        request.boolStreamData.data.extend(data)
        request.boolStreamData.SetInParent()
    response = job._job_manager_client.InsertInputStream(request)
    valid_errors = (
        MissingJobError,
        InvalidJobExecutionStatusError,
        UnknownInputStreamError,
    )
    _handle_job_manager_error(request, response, valid_errors)


def get_all_iterable_data_types(it):
    return set([type(e) for e in it])


def collection_has_type(
    collection: Collection, type_to_check: Type, include_subclasses: bool
) -> bool:
    if include_subclasses:
        return any([isinstance(i, type_to_check) for i in collection])
    else:
        return any([type(i) is type_to_check for i in collection])


def collection_has_type_bool(collection: Collection):
    return collection_has_type(collection, bool, False) or collection_has_type(
        collection, np.bool_, True
    )


def collection_has_type_int(collection: Collection):
    return collection_has_type(collection, int, False) or collection_has_type(
        collection, np.integer, True
    )


def collection_has_type_float(collection: Collection):
    return collection_has_type(collection, float, False) or collection_has_type(
        collection, np.floating, True
    )


def is_iter(x):
    try:
        iter(x)
    except TypeError:
        return False
    else:
        return True


def get_iterable_elements_datatype(it):
    if isinstance(it, np.ndarray):
        return type(it[0].item())
    elif is_iter(it):
        if len(get_all_iterable_data_types(it)) > 1:
            raise ValueError("Multiple datatypes encounterd in iterable object")
        if isinstance(it[0], np.generic):
            return type(it[0].item())
        else:
            return type(it[0])
    else:
        return None



from qm.octave._calibration_config import offset_amp
from qm.qua import (
    assign,
    for_,
    reset_phase,
    align,
    play,
    measure,
    dual_demod,
    if_,
    save,
    amp,
    update_correction,
    else_,
    while_,
    program,
    declare,
    fixed,
    declare_stream,
    wait,
    stream_processing,
)


def _generate_program(calibration_parameters):
    n_average = calibration_parameters["average_iterations"]
    iterations = calibration_parameters["iterations"]
    keep_on = calibration_parameters["keep_on"]
    initial_step_size = 2**-5

    def read_all_power():
        assign(avg_signal_power, 0)
        assign(avg_lo_power, 0)
        assign(avg_image_power, 0)

        with for_(index, 0, index < n_average, index + 1):
            reset_phase("IQmixer")
            reset_phase("signal_analyzer")
            reset_phase("lo_analyzer")
            reset_phase("image_analyzer")
            align(
                "IQmixer",
                "signal_analyzer",
                "lo_analyzer",
                "image_analyzer",
                "I_offset",
                "Q_offset",
            )

            play("calibration", "IQmixer")
            measure(
                "Analyze",
                "signal_analyzer",
                None,
                dual_demod.full("integW_cos", "out1", "integW_sin", "out2", I1),
                dual_demod.full("integW_cos", "out2", "integW_minus_sin", "out1", Q1),
            )
            measure(
                "Analyze",
                "lo_analyzer",
                None,
                dual_demod.full("integW_cos", "out1", "integW_sin", "out2", I2),
                dual_demod.full("integW_cos", "out2", "integW_minus_sin", "out1", Q2),
            )
            measure(
                "Analyze",
                "image_analyzer",
                None,
                dual_demod.full("integW_cos", "out1", "integW_sin", "out2", I3),
                dual_demod.full("integW_cos", "out2", "integW_minus_sin", "out1", Q3),
            )
            assign(
                avg_signal_power,
                avg_signal_power + (I1 * I1 + Q1 * Q1),
            )
            with if_(avg_signal_power < -1.0):
                save(avg_signal_power, "error")
            assign(
                avg_lo_power,
                avg_lo_power + (I2 * I2 + Q2 * Q2),
            )
            with if_(avg_lo_power < -1.0):
                save(avg_lo_power, "error")
            assign(
                avg_image_power,
                avg_image_power + (I3 * I3 + Q3 * Q3),
            )
            with if_(avg_image_power < -1.0):
                save(avg_image_power, "error")

        # assign(avg_signal_power, avg_signal_power * float(1 / n_average))
        # assign(avg_lo_power, avg_lo_power * float(1 / n_average))
        # assign(avg_image_power, avg_image_power * float(1 / n_average))

    def read_lo_and_image_power():
        assign(avg_lo_power, 0)
        assign(avg_image_power, 0)
        with for_(index, 0, index < n_average, index + 1):
            reset_phase("IQmixer")
            reset_phase("lo_analyzer")
            reset_phase("image_analyzer")
            align("IQmixer", "lo_analyzer", "image_analyzer", "I_offset", "Q_offset")
            play("calibration", "IQmixer")
            measure(
                "Analyze",
                "lo_analyzer",
                None,
                dual_demod.full("integW_cos", "out1", "integW_sin", "out2", I2),
                dual_demod.full("integW_cos", "out2", "integW_minus_sin", "out1", Q2),
            )
            measure(
                "Analyze",
                "image_analyzer",
                None,
                dual_demod.full("integW_cos", "out1", "integW_sin", "out2", I3),
                dual_demod.full("integW_cos", "out2", "integW_minus_sin", "out1", Q3),
            )
            assign(
                avg_lo_power,
                avg_lo_power + (I2 * I2 + Q2 * Q2),
            )
            with if_(avg_lo_power < -1.0):
                save(avg_lo_power, "error")
            assign(
                avg_image_power,
                avg_image_power + (I3 * I3 + Q3 * Q3),
            )
            with if_(avg_image_power < -1.0):
                save(avg_image_power, "error")

    def update_correction_g_phi():
        assign(s_mat, curr_phase)
        assign(s2, s_mat * s_mat)
        assign(c_mat, 1 + 1.5 * s2 - 3.125 * (s2 * s2))
        assign(g_mat_plus, 1 + curr_gain + curr_gain * curr_gain)
        assign(g_mat_minus, 1 - curr_gain + curr_gain * curr_gain)
        assign(c00, g_mat_minus * c_mat)
        assign(c01, g_mat_plus * s_mat)
        assign(c10, g_mat_minus * s_mat)
        assign(c11, g_mat_plus * c_mat)
        update_correction("IQmixer", c00, c01, c10, c11)

    def search_1_direction(which):
        with if_(lo_cont):
            # LO
            if which == "i_g":
                assign(curr_i0, curr_i0 + step_size_lo)
                play("DC_offset" * amp(step_size_lo), "I_offset")
            else:
                assign(curr_q0, curr_q0 + step_size_lo)
                play("DC_offset" * amp(step_size_lo), "Q_offset")
        with if_(image_cont):
            # image
            if which == "i_g":
                assign(curr_gain, curr_gain + step_size_image)
            else:
                assign(curr_phase, curr_phase + step_size_image)
            update_correction_g_phi()

        read_lo_and_image_power()
        with if_(lo_cont):
            # LO
            with if_(avg_lo_power < best_lo_power):  # Current direction is good
                assign(curr_step_lo, step_size_lo)
            with else_():  # Need to check other direction
                assign(curr_step_lo, -step_size_lo)
                if which == "i_g":
                    assign(curr_i0, curr_i0 + curr_step_lo)
                    play("DC_offset" * amp(curr_step_lo), "I_offset")
                else:
                    assign(curr_q0, curr_q0 + curr_step_lo)
                    play("DC_offset" * amp(curr_step_lo), "Q_offset")
                assign(avg_lo_power, best_lo_power)
        with if_(image_cont):
            # image
            with if_(avg_image_power < best_image_power):  # Current direction is good
                assign(curr_step_image, step_size_image)
            with else_():  # Need to check other direction
                assign(curr_step_image, -step_size_image)
                if which == "i_g":
                    assign(curr_gain, curr_gain + curr_step_image)
                else:
                    assign(curr_phase, curr_phase + curr_step_image)
                assign(avg_image_power, best_image_power)

        assign(move_lo_cont, lo_cont)
        assign(move_image_cont, image_cont)
        with while_((move_lo_cont | move_image_cont)):
            with if_(move_lo_cont):
                assign(best_lo_power, avg_lo_power)
                save(best_lo_power, lo_stream)
                if which == "i_g":
                    save(curr_i0, "i_track")
                    assign(curr_i0, curr_i0 + curr_step_lo)
                    play("DC_offset" * amp(curr_step_lo), "I_offset")
                else:
                    save(curr_q0, "q_track")
                    assign(curr_q0, curr_q0 + curr_step_lo)
                    play("DC_offset" * amp(curr_step_lo), "Q_offset")
            with if_(move_image_cont):
                assign(best_image_power, avg_image_power)
                save(best_image_power, image_stream)
                if which == "i_g":
                    save(curr_gain, "g_track")
                    assign(curr_gain, curr_gain + curr_step_image)
                else:
                    save(curr_phase, "phi_track")
                    assign(curr_phase, curr_phase + curr_step_image)
                update_correction_g_phi()
            assign(counter, counter + 1)
            read_lo_and_image_power()
            with if_(move_lo_cont):
                assign(move_lo_cont, avg_lo_power < best_lo_power)
            with if_(move_image_cont):
                assign(move_image_cont, avg_image_power < best_image_power)

        with if_(lo_cont):
            if which == "i_g":
                assign(curr_i0, curr_i0 - curr_step_lo)
                play("DC_offset" * amp(-curr_step_lo), "I_offset")
            else:
                assign(curr_q0, curr_q0 - curr_step_lo)
                play("DC_offset" * amp(-curr_step_lo), "Q_offset")
        with if_(image_cont):
            if which == "i_g":
                assign(curr_gain, curr_gain - curr_step_image)
            else:
                assign(curr_phase, curr_phase - curr_step_image)
            update_correction_g_phi()

    with program() as OPMix_minimize_line_search:
        index = declare(int)

        # variables for getting IQ data
        Q1 = declare(fixed)
        Q2 = declare(fixed)
        Q3 = declare(fixed)
        I1 = declare(fixed)
        I2 = declare(fixed)
        I3 = declare(fixed)

        # variables for current power
        avg_signal_power = declare(fixed)
        avg_lo_power = declare(fixed)
        avg_image_power = declare(fixed)

        # streams for power
        signal_stream = declare_stream()
        lo_stream = declare_stream()
        image_stream = declare_stream()

        # variables for LO leakage
        curr_i0 = declare(fixed, value=0)
        curr_q0 = declare(fixed, value=0)
        best_lo_power = declare(fixed, value=0)
        step_size_lo = declare(fixed, value=initial_step_size)
        curr_step_lo = declare(fixed, value=0)
        move_lo_cont = declare(bool, value=True)

        # variables for image leakage
        curr_gain = declare(fixed, value=0)
        curr_phase = declare(fixed, value=0)
        best_image_power = declare(fixed, value=0)
        step_size_image = declare(fixed, value=initial_step_size)
        curr_step_image = declare(fixed, value=0)
        move_image_cont = declare(bool, value=True)

        # variables for correction matrix
        g_mat_plus = declare(fixed)
        g_mat_minus = declare(fixed)
        s2 = declare(fixed)
        c_mat = declare(fixed)
        s_mat = declare(fixed)
        c00 = declare(fixed)
        c01 = declare(fixed)
        c10 = declare(fixed)
        c11 = declare(fixed)

        # iterations
        bool_lo_power_small = declare(bool)
        bool_lo_step_size_small = declare(bool)
        bool_image_power_small = declare(bool)
        bool_image_step_size_small = declare(bool)
        bool_too_many_iterations = declare(bool)
        lo_cont = declare(bool, value=True)
        image_cont = declare(bool, value=True)
        counter = declare(int, value=0)

        read_lo_and_image_power()
        save(curr_i0, "i_track")
        save(curr_q0, "q_track")
        save(avg_lo_power, lo_stream)
        save(curr_gain, "g_track")
        save(curr_phase, "phi_track")
        save(avg_image_power, image_stream)
        assign(best_lo_power, avg_lo_power)
        assign(best_image_power, avg_image_power)
        with while_((lo_cont | image_cont)):
            search_1_direction("i_g")
            search_1_direction("q_p")

            assign(step_size_lo, step_size_lo >> 1)  # Divide step size by 2
            assign(step_size_image, step_size_image >> 1)  # Divide step size by 2

            assign(bool_too_many_iterations, counter >= iterations)

            assign(bool_lo_power_small, avg_lo_power < 2**-27)
            assign(
                bool_lo_step_size_small, step_size_lo < 2**-16
            )  # opx fixed point resolution
            assign(
                lo_cont,
                ~(
                    bool_lo_power_small
                    | bool_lo_step_size_small
                    | bool_too_many_iterations
                ),
            )

            assign(bool_image_power_small, avg_image_power < 2**-27)
            assign(
                bool_image_step_size_small, step_size_image < 2**-16
            )  # opx fixed point resolution
            assign(
                image_cont,
                ~(
                    bool_image_power_small
                    | bool_image_step_size_small
                    | bool_too_many_iterations
                ),
            )

        save(bool_lo_power_small, "bool_lo_power_small")
        save(bool_lo_step_size_small, "bool_lo_step_size_small")
        save(bool_image_power_small, "bool_image_power_small")
        save(bool_image_step_size_small, "bool_image_step_size_small")
        save(bool_too_many_iterations, "bool_too_many_iterations")
        save(counter, "counter")

        read_all_power()
        save(avg_signal_power, signal_stream)
        save(avg_lo_power, lo_stream)
        save(avg_image_power, image_stream)
        save(c00, "c00")
        save(c01, "c01")
        save(c10, "c10")
        save(c11, "c11")

        if keep_on:
            with for_(index, 0, index < 1e6, index + 1):
                play("calibration", "IQmixer")
                wait(10, "I_offset")
                wait(10, "Q_offset")

        with stream_processing():
            lo_stream.save_all("lo")
            image_stream.save_all("image")
            signal_stream.save_all("signal")

    return OPMix_minimize_line_search


def _process_results(
    error,
    i_track,
    q_track,
    c00,
    c01,
    c10,
    c11,
):
    if len(error) > 0:
        raise RuntimeError("Readout pulse too long!")

    final_i = i_track[-1] * offset_amp
    final_q = q_track[-1] * offset_amp

    return [final_i, final_q], [c00[0], c01[0], c10[0], c11[0]]



import dataclasses
from enum import Enum, auto
from typing import Optional

try:
    from octave_sdk import (
        RFInputRFSource as sdk_RFInputRFSource,
        RFOutputMode as sdk_RFOutputMode,
        OctaveLOSource as sdk_OctaveLOSource,
        RFInputLOSource as sdk_RFInputLOSource,
        ClockType as sdk_ClockType,
        ClockFrequency as sdk_ClockFrequency,
        OctaveOutput as sdk_OctaveOutput,
    )
    from octave_sdk.octave import IFMode as sdk_IFMode
except ModuleNotFoundError:
    # Setting empty values for tests
    sdk_OctaveLOSource = None
    sdk_OctaveOutput = None
    sdk_RFInputRFSource = None
    sdk_RFOutputMode = None
    sdk_RFInputLOSource = None
    sdk_ClockType = None
    sdk_ClockFrequency = None


class RFInputRFSource(Enum):
    """ """

    RF_in = auto()
    Loopback_RF_out_1 = auto()
    Loopback_RF_out_2 = auto()
    Loopback_RF_out_3 = auto()
    Loopback_RF_out_4 = auto()
    Loopback_RF_out_5 = auto()
    Off = auto()

    def to_sdk(self):
        if self == RFInputRFSource.RF_in:
            return sdk_RFInputRFSource.RF_in
        elif self == RFInputRFSource.Off:
            return sdk_RFInputRFSource.Off
        elif self == RFInputRFSource.Loopback_RF_out_1:
            return sdk_RFInputRFSource.Loopback_RF_out_1
        elif self == RFInputRFSource.Loopback_RF_out_2:
            return sdk_RFInputRFSource.Loopback_RF_out_2
        elif self == RFInputRFSource.Loopback_RF_out_3:
            return sdk_RFInputRFSource.Loopback_RF_out_3
        elif self == RFInputRFSource.Loopback_RF_out_4:
            return sdk_RFInputRFSource.Loopback_RF_out_4
        elif self == RFInputRFSource.Loopback_RF_out_5:
            return sdk_RFInputRFSource.Loopback_RF_out_5
        else:
            return None

    @staticmethod
    def from_sdk(value):
        if value == sdk_RFInputRFSource.RF_in:
            return RFInputRFSource.RF_in
        elif value == sdk_RFInputRFSource.Off:
            return RFInputRFSource.Off
        elif value == sdk_RFInputRFSource.Loopback_RF_out_1:
            return RFInputRFSource.Loopback_RF_out_1
        elif value == sdk_RFInputRFSource.Loopback_RF_out_2:
            return RFInputRFSource.Loopback_RF_out_2
        elif value == sdk_RFInputRFSource.Loopback_RF_out_3:
            return RFInputRFSource.Loopback_RF_out_3
        elif value == sdk_RFInputRFSource.Loopback_RF_out_4:
            return RFInputRFSource.Loopback_RF_out_4
        elif value == sdk_RFInputRFSource.Loopback_RF_out_5:
            return RFInputRFSource.Loopback_RF_out_5
        else:
            raise ValueError(f"Clock type {value} is unknown")


class IFMode(Enum):
    """ """

    direct = auto()
    envelope = auto()
    mixer = auto()
    off = auto()

    def to_sdk(self):
        if self == IFMode.direct:
            return sdk_IFMode.direct
        elif self == IFMode.envelope:
            return sdk_IFMode.envelope
        elif self == IFMode.mixer:
            return sdk_IFMode.mixer
        elif self == IFMode.off:
            return sdk_IFMode.off
        else:
            return None

    @staticmethod
    def from_sdk(value):
        if value == sdk_IFMode.direct:
            return IFMode.direct
        elif value == sdk_IFMode.envelope:
            return IFMode.envelope
        elif value == sdk_IFMode.mixer:
            return IFMode.mixer
        elif value == sdk_IFMode.off:
            return IFMode.off
        else:
            raise ValueError(f"Clock type {value} is unknown")


class RFOutputMode(Enum):
    """ """

    trig_normal = auto()
    trig_inverse = auto()
    on = auto()
    off = auto()
    debug = auto()

    def to_sdk(self):
        if self == RFOutputMode.trig_normal:
            return sdk_RFOutputMode.trig_normal
        elif self == RFOutputMode.trig_inverse:
            return sdk_RFOutputMode.trig_inverse
        elif self == RFOutputMode.on:
            return sdk_RFOutputMode.on
        elif self == RFOutputMode.off:
            return sdk_RFOutputMode.off
        elif self == RFOutputMode.debug:
            return sdk_RFOutputMode.debug
        else:
            return None

    @staticmethod
    def from_sdk(value):
        if value == sdk_RFOutputMode.trig_normal:
            return RFOutputMode.trig_normal
        elif value == sdk_RFOutputMode.trig_inverse:
            return RFOutputMode.trig_inverse
        elif value == sdk_RFOutputMode.on:
            return RFOutputMode.on
        elif value == sdk_RFOutputMode.off:
            return RFOutputMode.off
        elif value == sdk_RFOutputMode.debug:
            return RFOutputMode.debug
        else:
            raise ValueError(f"Clock type {value} is unknown")


class OctaveLOSource(Enum):
    """input to the octave, uses it for output"""

    Off = auto()
    Internal = auto()
    Dmd1LO = auto()
    Dmd2LO = auto()
    LO1 = auto()
    LO2 = auto()
    LO3 = auto()
    LO4 = auto()
    LO5 = auto()
    RF_IN1 = auto()
    RF_IN2 = auto()
    IF1_LO_I = auto()
    IF1_LO_Q = auto()
    IF2_LO_I = auto()
    IF2_LO_Q = auto()

    def to_sdk(self):
        if self == OctaveLOSource.Off:
            return sdk_OctaveLOSource.Off
        elif self == OctaveLOSource.Internal:
            return sdk_OctaveLOSource.Internal
        elif self == OctaveLOSource.Dmd1LO:
            return sdk_OctaveLOSource.Dmd1LO
        elif self == OctaveLOSource.Dmd2LO:
            return sdk_OctaveLOSource.Dmd2LO
        elif self == OctaveLOSource.LO1:
            return sdk_OctaveLOSource.LO1
        elif self == OctaveLOSource.LO2:
            return sdk_OctaveLOSource.LO2
        elif self == OctaveLOSource.LO3:
            return sdk_OctaveLOSource.LO3
        elif self == OctaveLOSource.LO4:
            return sdk_OctaveLOSource.LO4
        elif self == OctaveLOSource.LO5:
            return sdk_OctaveLOSource.LO5
        elif self == OctaveLOSource.RF_IN1:
            return sdk_OctaveLOSource.RF_IN1
        elif self == OctaveLOSource.RF_IN2:
            return sdk_OctaveLOSource.RF_IN2
        elif self == OctaveLOSource.IF1_LO_I:
            return sdk_OctaveLOSource.IF1_LO_I
        elif self == OctaveLOSource.IF1_LO_Q:
            return sdk_OctaveLOSource.IF1_LO_Q
        elif self == OctaveLOSource.IF2_LO_I:
            return sdk_OctaveLOSource.IF2_LO_I
        elif self == OctaveLOSource.IF2_LO_Q:
            return sdk_OctaveLOSource.IF2_LO_Q
        else:
            return None

    @staticmethod
    def from_sdk(value):
        if value == sdk_OctaveLOSource.Off:
            return OctaveLOSource.Off
        elif value == sdk_OctaveLOSource.Internal:
            return OctaveLOSource.Internal
        elif value == sdk_OctaveLOSource.Dmd1LO:
            return OctaveLOSource.Dmd1LO
        elif value == sdk_OctaveLOSource.Dmd2LO:
            return OctaveLOSource.Dmd2LO
        elif value == sdk_OctaveLOSource.LO1:
            return OctaveLOSource.LO1
        elif value == sdk_OctaveLOSource.LO2:
            return OctaveLOSource.LO2
        elif value == sdk_OctaveLOSource.LO3:
            return OctaveLOSource.LO3
        elif value == sdk_OctaveLOSource.LO4:
            return OctaveLOSource.LO4
        elif value == sdk_OctaveLOSource.LO5:
            return OctaveLOSource.LO5
        elif value == sdk_OctaveLOSource.RF_IN1:
            return OctaveLOSource.RF_IN1
        elif value == sdk_OctaveLOSource.RF_IN2:
            return OctaveLOSource.RF_IN2
        elif value == sdk_OctaveLOSource.IF1_LO_I:
            return OctaveLOSource.IF1_LO_I
        elif value == sdk_OctaveLOSource.IF1_LO_Q:
            return OctaveLOSource.IF1_LO_Q
        elif value == sdk_OctaveLOSource.IF2_LO_I:
            return OctaveLOSource.IF2_LO_I
        elif value == sdk_OctaveLOSource.IF2_LO_Q:
            return OctaveLOSource.IF2_LO_Q
        else:
            raise ValueError(f"Clock type {value} is unknown")


class RFInputLOSource(Enum):
    """
    input to octave, uses it for downconversion, the lo
    extending the OctaveLOSourceInput
    """

    Off = auto()
    Internal = auto()
    Dmd1LO = auto()
    Dmd2LO = auto()
    LO1 = auto()
    LO2 = auto()
    LO3 = auto()
    LO4 = auto()
    LO5 = auto()
    Analyzer = auto()
    RFOutput1_LO = auto()
    RFOutput2_LO = auto()
    RFOutput3_LO = auto()
    RFOutput4_LO = auto()
    RFOutput5_LO = auto()

    def to_sdk(self):
        if self == RFInputLOSource.Off:
            return sdk_RFInputLOSource.Off
        elif self == RFInputLOSource.Internal:
            return sdk_RFInputLOSource.Internal
        elif self == RFInputLOSource.Dmd1LO:
            return sdk_RFInputLOSource.Dmd1LO
        elif self == RFInputLOSource.Dmd2LO:
            return sdk_RFInputLOSource.Dmd2LO
        elif self == RFInputLOSource.LO1:
            return sdk_RFInputLOSource.LO1
        elif self == RFInputLOSource.LO2:
            return sdk_RFInputLOSource.LO2
        elif self == RFInputLOSource.LO3:
            return sdk_RFInputLOSource.LO3
        elif self == RFInputLOSource.LO4:
            return sdk_RFInputLOSource.LO4
        elif self == RFInputLOSource.LO5:
            return sdk_RFInputLOSource.LO5
        elif self == RFInputLOSource.Analyzer:
            return sdk_RFInputLOSource.Analyzer
        elif self == RFInputLOSource.RFOutput1_LO:
            return sdk_RFInputLOSource.RFOutput1_LO
        elif self == RFInputLOSource.RFOutput2_LO:
            return sdk_RFInputLOSource.RFOutput2_LO
        elif self == RFInputLOSource.RFOutput3_LO:
            return sdk_RFInputLOSource.RFOutput3_LO
        elif self == RFInputLOSource.RFOutput4_LO:
            return sdk_RFInputLOSource.RFOutput4_LO
        elif self == RFInputLOSource.RFOutput5_LO:
            return sdk_RFInputLOSource.RFOutput5_LO
        else:
            return None

    @staticmethod
    def from_sdk(value):
        if value == sdk_RFInputLOSource.Off:
            return RFInputLOSource.Off
        elif value == sdk_RFInputLOSource.Internal:
            return RFInputLOSource.Internal
        elif value == sdk_RFInputLOSource.Dmd1LO:
            return RFInputLOSource.Dmd1LO
        elif value == sdk_RFInputLOSource.Dmd2LO:
            return RFInputLOSource.Dmd2LO
        elif value == sdk_RFInputLOSource.LO1:
            return RFInputLOSource.LO1
        elif value == sdk_RFInputLOSource.LO2:
            return RFInputLOSource.LO2
        elif value == sdk_RFInputLOSource.LO3:
            return RFInputLOSource.LO3
        elif value == sdk_RFInputLOSource.LO4:
            return RFInputLOSource.LO4
        elif value == sdk_RFInputLOSource.LO5:
            return RFInputLOSource.LO5
        elif value == sdk_RFInputLOSource.Analyzer:
            return RFInputLOSource.Analyzer
        elif value == sdk_RFInputLOSource.RFOutput1_LO:
            return RFInputLOSource.RFOutput1_LO
        elif value == sdk_RFInputLOSource.RFOutput2_LO:
            return RFInputLOSource.RFOutput2_LO
        elif value == sdk_RFInputLOSource.RFOutput3_LO:
            return RFInputLOSource.RFOutput3_LO
        elif value == sdk_RFInputLOSource.RFOutput4_LO:
            return RFInputLOSource.RFOutput4_LO
        elif value == sdk_RFInputLOSource.RFOutput5_LO:
            return RFInputLOSource.RFOutput5_LO
        else:
            raise ValueError(f"Clock type {value} is unknown")


class ClockType(Enum):
    """ """

    Internal = auto()
    External = auto()
    Buffered = auto()  # for opt

    def to_sdk(self):
        if self == ClockType.Internal:
            return sdk_ClockType.Internal
        elif self == ClockType.Buffered:
            return sdk_ClockType.Buffered
        elif self == ClockType.External:
            return sdk_ClockType.External
        else:
            return None

    @staticmethod
    def from_sdk(value):
        if value == sdk_ClockType.Internal:
            return ClockType.Internal
        elif value == sdk_ClockType.Buffered:
            return ClockType.Buffered
        elif value == sdk_ClockType.External:
            return ClockType.External
        else:
            raise ValueError(f"Clock type {value} is unknown")


class ClockFrequency(Enum):
    """ """

    MHZ_10 = auto()
    MHZ_100 = auto()
    MHZ_1000 = auto()

    def to_sdk(self):
        if self == ClockFrequency.MHZ_10:
            return sdk_ClockFrequency.MHZ_10
        elif self == ClockFrequency.MHZ_100:
            return sdk_ClockFrequency.MHZ_100
        elif self == ClockFrequency.MHZ_1000:
            return sdk_ClockFrequency.MHZ_1000
        else:
            return None

    @staticmethod
    def from_sdk(value):
        if value == sdk_ClockFrequency.MHZ_10:
            return ClockFrequency.MHZ_10
        elif value == sdk_ClockFrequency.MHZ_100:
            return ClockFrequency.MHZ_100
        elif value == sdk_ClockFrequency.MHZ_1000:
            return ClockFrequency.MHZ_1000
        else:
            raise ValueError(f"Clock type {value} is unknown")


class OctaveOutput(Enum):
    Synth1 = auto()
    Synth2 = auto()
    Synth3 = auto()
    Synth4 = auto()
    Synth5 = auto()
    RF1 = auto()
    RF2 = auto()
    RF3 = auto()
    RF4 = auto()
    RF5 = auto()
    IF_OUT1 = auto()
    IF_OUT2 = auto()
    Dig1 = auto()
    Dig2 = auto()
    Dig3 = auto()

    def to_sdk(self):
        if self == OctaveOutput.Synth1:
            return sdk_OctaveOutput.Synth1
        elif self == OctaveOutput.Synth2:
            return sdk_OctaveOutput.Synth2
        elif self == OctaveOutput.Synth3:
            return sdk_OctaveOutput.Synth3
        elif self == OctaveOutput.Synth4:
            return sdk_OctaveOutput.Synth4
        elif self == OctaveOutput.Synth5:
            return sdk_OctaveOutput.Synth5
        elif self == OctaveOutput.RF1:
            return sdk_OctaveOutput.RF1
        elif self == OctaveOutput.RF2:
            return sdk_OctaveOutput.RF2
        elif self == OctaveOutput.RF3:
            return sdk_OctaveOutput.RF3
        elif self == OctaveOutput.RF4:
            return sdk_OctaveOutput.RF4
        elif self == OctaveOutput.RF5:
            return sdk_OctaveOutput.RF5
        elif self == OctaveOutput.IF_OUT1:
            return sdk_OctaveOutput.IF_OUT1
        elif self == OctaveOutput.IF_OUT2:
            return sdk_OctaveOutput.IF_OUT2
        elif self == OctaveOutput.Dig1:
            return sdk_OctaveOutput.Dig1
        elif self == OctaveOutput.Dig2:
            return sdk_OctaveOutput.Dig2
        elif self == OctaveOutput.Dig3:
            return sdk_OctaveOutput.Dig3
        else:
            return None

    @staticmethod
    def from_sdk(value):
        if value == sdk_OctaveOutput.Synth1:
            return OctaveOutput.Synth1
        elif value == sdk_OctaveOutput.Synth2:
            return OctaveOutput.Synth2
        elif value == sdk_OctaveOutput.Synth3:
            return OctaveOutput.Synth3
        elif value == sdk_OctaveOutput.Synth4:
            return OctaveOutput.Synth4
        elif value == sdk_OctaveOutput.Synth5:
            return OctaveOutput.Synth5
        elif value == sdk_OctaveOutput.RF1:
            return OctaveOutput.RF1
        elif value == sdk_OctaveOutput.RF2:
            return OctaveOutput.RF2
        elif value == sdk_OctaveOutput.RF3:
            return OctaveOutput.RF3
        elif value == sdk_OctaveOutput.RF4:
            return OctaveOutput.RF4
        elif value == sdk_OctaveOutput.RF5:
            return OctaveOutput.RF5
        elif value == sdk_OctaveOutput.IF_OUT1:
            return OctaveOutput.IF_OUT1
        elif value == sdk_OctaveOutput.IF_OUT2:
            return OctaveOutput.IF_OUT2
        elif value == sdk_OctaveOutput.Dig1:
            return OctaveOutput.Dig1
        elif value == sdk_OctaveOutput.Dig2:
            return OctaveOutput.Dig2
        elif value == sdk_OctaveOutput.Dig3:
            return OctaveOutput.Dig3
        else:
            raise ValueError(f"Clock type {value} is unknown")


@dataclasses.dataclass
class ClockInfo:
    clock_type: ClockType
    frequency: Optional[ClockFrequency]



import dataclasses
import re
from typing import Dict, Optional, Tuple, TypeVar, Generic, List, Union

from qm.octave import OctaveOutput, OctaveLOSource
from qm.octave.calibration_db import CalibrationDB
from qm.logger import logger

ConnectionMapping = Dict[Tuple[str, int], Tuple[str, str]]

DEFAULT_CONNECTIONS = {
    1: "I1",
    2: "Q1",
    3: "I2",
    4: "Q2",
    5: "I3",
    6: "Q3",
    7: "I4",
    8: "Q4",
    9: "I5",
    10: "Q5",
}


@dataclasses.dataclass
class ConnectionInfo:
    host: str
    port: int

    def __hash__(self):
        return hash((self.host, self.port))


PortType = TypeVar("PortType", OctaveOutput, OctaveLOSource)


@dataclasses.dataclass
class LoopbackInfo(Generic[PortType]):
    name: str
    octave_port: PortType

    def __hash__(self):
        return hash((self.name, self.octave_port))


def _convert_octave_port_to_number(port: str) -> int:
    if port == "I1":
        return 1
    elif port == "Q1":
        return 1
    elif port == "I2":
        return 2
    elif port == "Q2":
        return 2
    elif port == "I3":
        return 3
    elif port == "Q3":
        return 3
    elif port == "I4":
        return 4
    elif port == "Q4":
        return 4
    elif port == "I5":
        return 5
    elif port == "Q5":
        return 5


def _convert_number_to_octave_port(name: str, port: int) -> List:
    return [(name, f"I{port}"), (name, f"Q{port}")]


class QmOctaveConfig:
    def __init__(self, **kwargs):
        self._devices: Dict[str, ConnectionInfo] = {}
        self._calibration_db_path: Optional[str] = None
        self._loopbacks: Dict[
            LoopbackInfo[OctaveLOSource], LoopbackInfo[OctaveOutput]
        ] = {}
        self._opx_octave_port_mapping: ConnectionMapping = {}
        self._calibration_db: Optional[CalibrationDB] = None
        self._fan = None
        if "fan" in kwargs:
            self._fan = kwargs.get("fan")

    @property
    def fan(self):
        return self._fan

    def add_device_info(self, name: str, host: str, port: int):
        """
        Sets the octave info  - the IP address can be either the router ip or the actual ip
        depends on the installation configuration (cluster or standalone mode)

        :param name: str
        :param host: str
        :param port: int
        """
        # can be either the router ip or the actual ip (depends on what we have)
        self._devices[name] = ConnectionInfo(host, port)

    def get_devices(self) -> Dict[str, ConnectionInfo]:
        return self._devices

    def set_calibration_db(self, path):
        """
        Sets the path to the calibration DB

        :param path: str
        """
        self._calibration_db_path = path

    @staticmethod
    def _check_connection_format(connection: Tuple[str, Union[int, str]]) -> bool:
        return (
            isinstance(connection, tuple)
            and len(connection) == 2
            and isinstance(connection[0], str)
        )

    def set_opx_octave_mapping(self, mappings: List[Tuple[str, str]]):
        """
        Sets the default port mapping for each `opx, octave` names
        """
        if self._opx_octave_port_mapping:
            logger.warning(
                "Setting opx-octave default port mapping, "
                "your configured mapping will be overridden!"
            )

        for opx_name, octave_name in mappings:
            self._opx_octave_port_mapping.update(
                self.get_default_opx_octave_port_mapping(opx_name, octave_name)
            )

    def add_opx_octave_port_mapping(self, connections: ConnectionMapping):
        """
        Adds port mapping which is different than the default one. should be in the form:
        Dict[Tuple["con1", 1], Tuple["octave1", "I1"]
        ...
        ...
        ]

        :param connections: ConnectionMapping
        """
        # ("con1", 1): ("octave1", "I1")
        # validate structure:
        for opx_connection, octave_connection in connections.items():
            if not (
                self._check_connection_format(opx_connection)
                and isinstance(opx_connection[1], int)
            ):
                raise ValueError(
                    f"key {opx_connection} is not according to format "
                    f'("con_name", "port_index")'
                )
            pattern = re.compile("([IQ])([12345])")
            if not (
                self._check_connection_format(octave_connection)
                and isinstance(octave_connection[1], str)
                and pattern.match(octave_connection[1]) is not None
            ):
                raise ValueError(
                    f"value {octave_connection} is not according to format "
                    f'("octave_name", "octave_port")'
                )

        self._opx_octave_port_mapping.update(connections)

    def get_opx_octave_port_mapping(self) -> ConnectionMapping:
        """
        Get the configured opx-octave connections
        """
        if self._opx_octave_port_mapping:
            return self._opx_octave_port_mapping
        else:
            raise RuntimeError(
                "No opx-octave mapping available, "
                "please run 'set_opx_octave_mapping' or "
                "'add_opx_octave_port_mapping' "
                "before attempting to read them."
            )

    @staticmethod
    def get_default_opx_octave_port_mapping(
        controller_name: str, octave_name: str
    ) -> ConnectionMapping:
        """
        Get the default opt-octave connections
        """
        default_connections = {}
        for opx_port, octave_port in DEFAULT_CONNECTIONS.items():
            default_connections[(controller_name, opx_port)] = (
                octave_name,
                octave_port,
            )
        return default_connections

    @property
    def calibration_db(self) -> CalibrationDB:
        if (
            self._calibration_db is None
            or self._calibration_db_path == self._calibration_db.file_path
        ) and self._calibration_db_path is not None:
            self._calibration_db = CalibrationDB(self._calibration_db_path)
        return self._calibration_db

    def add_lo_loopback(
        self,
        octave_output_name: str,
        octave_output_port: OctaveOutput,
        octave_input_name: str,
        octave_input_port: OctaveLOSource,
    ):
        """
        Adds a loopback between an OctaveOutput and an OctaveLOSource
        """

        if octave_output_name != octave_input_name:
            raise ValueError(
                "lo loopback between different octave devices are not supported"
            )
        loop_back_source = LoopbackInfo(octave_output_name, octave_output_port)
        loop_back_destination = LoopbackInfo(octave_input_name, octave_input_port)
        self._loopbacks[loop_back_destination] = loop_back_source

    def get_lo_loopbacks_by_octave(
        self, octave_name: str
    ) -> Dict[OctaveLOSource, OctaveOutput]:
        """
        Gets a list of all loop backs by octave name
        """

        result = {}
        for destination, source in self._loopbacks.items():
            if destination.name == octave_name:
                result[destination.octave_port] = source.octave_port
        return result

    def get_opx_iq_ports(self, octave_output_port: Tuple[str, int]):
        conns = self.get_opx_octave_port_mapping()
        inv_conns = {v: k for k, v in conns.items()}
        octave_input_port_i, octave_input_port_q = _convert_number_to_octave_port(
            octave_output_port[0], octave_output_port[1]
        )
        if octave_input_port_i not in inv_conns:
            octave_name, octave_port = octave_input_port_i
            raise KeyError(
                f"Could not find opx connections to port "
                f"'{octave_port}' of octave '{octave_name}'"
            )

        if octave_input_port_q not in inv_conns:
            octave_name, octave_port = octave_input_port_q
            raise KeyError(
                f"Could not find opx connections to port "
                f"'{octave_port}' of octave '{octave_name}'"
            )
        return [inv_conns[octave_input_port_i], inv_conns[octave_input_port_q]]



import os
from dataclasses import dataclass, asdict
from pathlib import Path
from typing import List, Dict, Union, Any, Optional

from tinydb import TinyDB, Query

from qm.logger import logger
from qm.octave.enums import OctaveLOSource


@dataclass
class CalibrationResult:
    correction: List[float]
    i_offset: float
    q_offset: float
    lo_frequency: float
    if_frequency: float
    temperature: float
    mixer_id: str
    lo_source = OctaveLOSource
    optimizer_parameters: Dict[str, Any]


class CalibrationDB:
    def __init__(self, path: Union[str, Path]) -> None:
        super().__init__()
        file_name = "calibration_db.json"
        self._file_path = os.path.join(path, file_name)
        self._db = TinyDB(self._file_path)

    @property
    def file_path(self):
        return self._file_path

    def update_calibration_data(
        self, data: Union[CalibrationResult, List[CalibrationResult]]
    ):
        if isinstance(data, CalibrationResult):
            data = [data]

        query = Query()
        for result in data:
            self._db.upsert(
                asdict(result),
                (query.mixer_id == result.mixer_id)
                & (query.lo_frequency == result.lo_frequency)
                & (query.if_frequency == result.if_frequency),
            )

    def get(self, mixer, lo_freq, if_freq) -> Optional[CalibrationResult]:
        query = Query()
        doc = self._db.get(
            (query.mixer_id == mixer)
            & (query.lo_frequency == lo_freq)
            & (query.if_frequency == if_freq),
        )
        if doc is not None:
            return self._doc_to_result(doc)
        else:
            raise AttributeError(
                f"calibration for {mixer}, lo {lo_freq}, if {if_freq} was not found"
            )

    def get_or_none(self, mixer, lo_freq, if_freq) -> Optional[CalibrationResult]:
        query = Query()
        doc = self._db.get(
            (query.mixer_id == mixer)
            & (query.lo_frequency == lo_freq)
            & (query.if_frequency == if_freq)
        )
        if doc is not None:
            return self._doc_to_result(doc)
        else:
            return None

    def get_for_lo_frequency(self, mixer, lo_freq) -> List[CalibrationResult]:
        query = Query()
        table = self._db.search(
            (query.mixer_id == mixer) & (query.lo_frequency == lo_freq)
        )
        return [self._doc_to_result(doc) for doc in table]

    def get_all(self, mixer) -> List[CalibrationResult]:
        l = self.get_all_or_default(mixer)
        if len(l) > 0:
            return l
        else:
            raise AttributeError(f"calibration for {mixer} was not found")

    def get_all_or_default(self, mixer) -> Optional[List[CalibrationResult]]:
        query = Query()
        table = self._db.search(query.mixer_id == mixer)
        l = [self._doc_to_result(doc) for doc in table]
        return l

    def __getitem__(self, item):
        return self.get(*item)

    @staticmethod
    def _doc_to_result(doc):
        return CalibrationResult(
            correction=doc["correction"],
            i_offset=doc["i_offset"],
            q_offset=doc["q_offset"],
            if_frequency=doc["if_frequency"],
            lo_frequency=doc["lo_frequency"],
            temperature=doc["temperature"],
            optimizer_parameters=doc["optimizer_parameters"],
            mixer_id=doc["mixer_id"],
        )

    def __del__(self):
        self._db.close()


def octave_output_mixer_name(octave: str, port: int) -> str:
    return f"octave_{octave}_{port}"


def load_from_calibration_db(config: Dict, calibration_db: CalibrationDB) -> None:
    if calibration_db is None:
        return
    for mixer in config["mixers"].keys():
        cal_list = calibration_db.get_all_or_default(mixer)
        cal_dict = {(v.lo_frequency, v.if_frequency): v.correction for v in cal_list}
        if len(cal_list) > 0:
            old_values = config["mixers"][mixer]
            new_values = [
                {
                    "intermediate_frequency": int(v.if_frequency),
                    "lo_frequency": int(v.lo_frequency),
                    "correction": v.correction,
                }
                for v in cal_list
            ]

            for v in old_values:
                if (v["lo_frequency"], v["intermediate_frequency"]) in cal_dict:
                    pass
                else:
                    new_values.append(v)
                    logger.debug(
                        f"Could not find calibration value for lo frequency"
                        f" {v['lo_frequency']} and intermediate_frequency"
                        f" {v['intermediate_frequency']}"
                    )

            config["mixers"][mixer] = new_values

            # changing offsets here because we know the mixer is octave
            for element in config["elements"].values():
                if (
                    "mixInputs" in element
                    and element["mixInputs"]["mixer"] == mixer
                    and "intermediate_frequency" in element
                ):
                    if_freq = element["intermediate_frequency"]
                    lo_freq = element["mixInputs"]["lo_frequency"]
                    i_port = element["mixInputs"]["I"]
                    q_port = element["mixInputs"]["Q"]
                    result = calibration_db.get_or_none(mixer, lo_freq, if_freq)
                    if result is not None:
                        config["controllers"][i_port[0]]["analog_outputs"][i_port[1]][
                            "offset"
                        ] = result.i_offset
                        config["controllers"][q_port[0]]["analog_outputs"][q_port[1]][
                            "offset"
                        ] = result.q_offset



from qm import QuantumMachine
from qm.octave._calibration_config import (
    _prep_config,
    _get_frequencies,
)
from qm.octave.calibration_db import (
    CalibrationResult,
    octave_output_mixer_name,
)
from qm.octave.enums import (
    OctaveLOSource,
    RFInputRFSource,
    ClockType,
    ClockFrequency,
    ClockInfo,
    RFOutputMode,
    RFInputLOSource,
    IFMode,
)
from qm.octave.octave_config import _convert_octave_port_to_number, QmOctaveConfig

from typing import Dict, Tuple, Optional, List, TYPE_CHECKING

if TYPE_CHECKING:
    from qm.QuantumMachinesManager import QuantumMachinesManager

try:
    from octave_sdk import Octave

    OCTAVE_SDK_LOADED = True

except ModuleNotFoundError:
    # Error handling
    OCTAVE_SDK_LOADED = False
    Octave = None


class OctaveSDKException(Exception):
    pass


class SetFrequencyException(Exception):
    pass


def _run_compiled(compiled_id: str, qm: QuantumMachine):
    from qm.octave._calibration_program import _process_results

    pending_job = qm.queue.add_compiled(compiled_id)
    job = pending_job.wait_for_execution()
    all_done = job.result_handles.wait_for_all_values(timeout=30)
    if not all_done:
        _calibration_failed_error()
    res = _get_results(
        [
            "error",
            "i_track",
            "q_track",
            "c00",
            "c01",
            "c10",
            "c11",
        ],
        job,
    )
    # g_track,
    # phi_track,
    # bool_lo_power_small,
    # bool_lo_step_size_small,
    # bool_too_many_iterations,
    # bool_image_power_small,
    # bool_image_step_size_small,
    # final_g = g_track[-1]
    # final_phi = phi_track[-1]
    # import numpy as np
    # lo = lo
    # if lo[-1] == 0:
    #     lo[-1] = 2 ** -28
    # image = image
    # if image[-1] == 0:
    #     image[-1] = 2 ** -28
    # signal = signal
    # if signal[-1] < 0:
    #     signal[-1] = 8 - signal[-1]
    # exit_lo = (
    #     1 * bool_lo_power_small
    #     + 2 * bool_lo_step_size_small
    #     + 4 * bool_too_many_iterations
    # )
    # print("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%")
    # print(
    #     f"LO suppression is {10 * np.log10(signal[-1] / lo[-1]):.4f} dBc at (I,"
    #     f"Q) = ({final_i:.4f},{final_q:.4f}) "
    # )
    # print(exit_lo)
    # #
    # exit_image = (
    #     1 * bool_image_power_small
    #     + 2 * bool_image_step_size_small
    #     + 4 * bool_too_many_iterations
    # )
    # print(
    #     f"Image suppression is {10 * np.log10(signal[-1] / image[-1]):.4f} dBc "
    #     f"at (g,phi) = ({final_g:.4f},{final_phi:.4f}) "
    # )
    # print(exit_image)
    # print("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%")
    return _process_results(
        res["error"],
        res["i_track"],
        res["q_track"],
        res["c00"],
        res["c01"],
        res["c10"],
        res["c11"],
    )


def _calibration_failed_error(extra: str = None):
    raise Exception(
        f"There was a problem during calibration. please check opx is on"
        f" and valid. If this continues, please contact QM support{extra}"
    )


def _get_results(names, job):
    res = {}
    for name in names:
        if name not in job.result_handles:
            _calibration_failed_error(f"\nerror in result {name}")
        res[name] = job.result_handles.get(name).fetch_all(flat_struct=True)
    return res


def _convert_rf_output_index_to_input_source(rf_output) -> RFInputRFSource:
    if rf_output == 1:
        return RFInputRFSource.Loopback_RF_out_1
    elif rf_output == 2:
        return RFInputRFSource.Loopback_RF_out_2
    elif rf_output == 3:
        return RFInputRFSource.Loopback_RF_out_3
    elif rf_output == 4:
        return RFInputRFSource.Loopback_RF_out_4
    elif rf_output == 5:
        return RFInputRFSource.Loopback_RF_out_5


class OctaveManager:
    def __init__(
        self,
        config: Optional[QmOctaveConfig] = None,
        qmm: Optional["QuantumMachinesManager"] = None,
    ) -> None:

        self._octave_config: QmOctaveConfig
        self._qmm = qmm

        self._check_and_set_config(config)
        self._initialize()

    def _check_and_set_config(self, config):
        if config and isinstance(config, QmOctaveConfig):
            self._octave_config: QmOctaveConfig = config
        elif config is None:
            self._octave_config: QmOctaveConfig = QmOctaveConfig()
        else:
            raise TypeError("config must be of type QmOctaveConfig")

    def _initialize(self):
        self._octave_clients: Dict[str, Octave] = {}
        devices = self._octave_config.get_devices()

        if devices is not None and len(devices) > 0:
            if not OCTAVE_SDK_LOADED:
                raise OctaveSDKException("Octave sdk is not installed")
            for device_name, connection_info in devices.items():
                loop_backs = self._octave_config.get_lo_loopbacks_by_octave(device_name)
                loop_backs = {
                    (input_port.to_sdk()): (output_port.to_sdk())
                    for input_port, output_port in loop_backs.items()
                }
                self._octave_clients[device_name] = Octave(
                    host=connection_info.host,
                    port=connection_info.port,
                    port_mapping=loop_backs,
                    octave_name=device_name,
                    fan=self._octave_config.fan,
                )

    def set_octave_configuration(self, config: QmOctaveConfig):
        """

        :param config:
        """
        self._check_and_set_config(config)
        self._initialize()

    def get_output_port(
        self,
        opx_i_port: Tuple[str, int],
        opx_q_port: Tuple[str, int],
    ) -> Tuple[str, int]:
        # check both ports are going to the same mixer
        connections = self._octave_config.get_opx_octave_port_mapping()
        i_octave_port = connections[opx_i_port]
        q_octave_port = connections[opx_q_port]
        if i_octave_port[0] != q_octave_port[0]:
            raise Exception("I and Q are not connected to the same octave")
        if i_octave_port[1][-1] != q_octave_port[1][-1]:
            raise Exception("I and Q are not connected to the same octave input")

        return i_octave_port[0], _convert_octave_port_to_number(i_octave_port[1])

    def _get_client(self, octave_port):
        octave = self._octave_clients[octave_port[0]]
        return octave

    def restore_default_state(self, octave_name):
        self._octave_clients[octave_name].restore_default_state()

    def set_clock(
        self,
        octave_name: str,
        clock_type: ClockType,
        frequency: Optional[ClockFrequency],
    ):

        """
        This function will set the octave clock type - internal, external or buffered.
        It can also set the clock frequency - 10, 100 or 1000 MHz

        :param octave_name: str
        :param clock_type: ClockType
        :param frequency: ClockFrequency
        :return:
        """
        self._octave_clients[octave_name].set_clock(
            clock_type.to_sdk(), frequency.to_sdk()
        )
        self._octave_clients[octave_name].save_default_state(only_clock=True)

    def get_clock(self, octave_name: str) -> ClockInfo:
        """
        Return the octave clock type and frequency

        :param octave_name: str
        :return: ClockInfo
        """
        sdk_clock = self._octave_clients[octave_name].get_clock()
        return ClockInfo(
            ClockType.from_sdk(sdk_clock.clock_type),
            ClockFrequency.from_sdk(sdk_clock.frequency),
        )

    def set_lo_frequency(
        self,
        octave_output_port: Tuple[str, int],
        lo_frequency: float,
        lo_source: OctaveLOSource = OctaveLOSource.Internal,
        set_source: bool = True,
    ):
        """
        Sets the LO frequency of the synthesizer associated to element

        :param octave_output_port:
        :param lo_frequency:
        :param lo_source:
        :param set_source:
        """
        octave = self._get_client(octave_output_port)
        octave_name, port_index = octave_output_port

        loop_backs = self._octave_config.get_lo_loopbacks_by_octave(octave_name)

        if lo_source != OctaveLOSource.Internal and lo_source not in loop_backs:
            raise SetFrequencyException(
                f"Cannot set frequency to an external lo source" f" {lo_source.name}"
            )

        if set_source:
            octave.rf_outputs[port_index].set_lo_source(lo_source.to_sdk())

        octave.rf_outputs[port_index].set_lo_frequency(lo_source.to_sdk(), lo_frequency)
        # octave.save_default_state()

    def set_lo_source(
        self, octave_output_port: Tuple[str, int], lo_port: OctaveLOSource
    ):
        """
        Sets the source of LO going to the upconverter associated with element.

        :param octave_output_port:
        :param lo_port:
        """
        octave = self._get_client(octave_output_port)
        octave.rf_outputs[octave_output_port[1]].set_lo_source(lo_port.to_sdk())
        # octave.save_default_state()

    def set_rf_output_mode(
        self, octave_output_port: Tuple[str, int], switch_mode: RFOutputMode
    ):
        """
        Configures the output switch of the upconverter associated to element.
        switch_mode can be either: 'always_on', 'always_off', 'normal' or 'inverted'
        When in 'normal' mode a high trigger will turn the switch on and a low
        trigger will turn it off
        When in 'inverted' mode a high trigger will turn the switch off and a low
        trigger will turn it on
        When in 'always_on' the switch will be permanently on. When in 'always_off'
        mode the switch will be permanently off.

        :param octave_output_port:
        :param switch_mode:
        """
        octave = self._get_client(octave_output_port)
        octave.rf_outputs[octave_output_port[1]].set_output(switch_mode.to_sdk())
        # octave.save_default_state()

    def set_rf_output_gain(
        self,
        octave_output_port: Tuple[str, int],
        gain_in_db: float,
        lo_frequency: Optional[float] = None,
    ):
        """
        Sets the RF output gain for the upconverter associated with element.
        if no lo_frequency is given, and lo source is internal, will use the
        internal frequency

        :param octave_output_port:
        :param gain_in_db:
        :param lo_frequency:
        """
        octave = self._get_client(octave_output_port)
        octave.rf_outputs[octave_output_port[1]].set_gain(gain_in_db, lo_frequency)
        # octave.save_default_state()

    def set_downconversion_lo_source(
        self,
        octave_input_port: Tuple[str, int],
        lo_source: RFInputLOSource,
        lo_frequency: Optional[float] = None,
        disable_warning: Optional[bool] = False,
    ):
        """
        Sets the LO source for the downconverters.

        :param octave_input_port:
        :param lo_source:
        :param lo_frequency:
        :param disable_warning:
        """
        octave = self._get_client(octave_input_port)
        octave.rf_inputs[octave_input_port[1]].set_lo_source(lo_source.to_sdk())
        octave.rf_inputs[octave_input_port[1]].set_rf_source(
            RFInputRFSource.RF_in.to_sdk()
        )
        internal = (
            lo_source == RFInputLOSource.Internal
            or lo_source == RFInputLOSource.Analyzer
        )
        if lo_frequency is not None and internal:
            octave.rf_inputs[octave_input_port[1]].set_lo_frequency(
                source_name=lo_source.to_sdk(), frequency=lo_frequency
            )
        # octave.save_default_state()

    def set_downconversion_if_mode(
        self,
        octave_input_port: Tuple[str, int],
        if_mode_i: IFMode = IFMode.direct,
        if_mode_q: IFMode = IFMode.direct,
        disable_warning: Optional[bool] = False,
    ):
        """
        Sets the IF downconversion stage.
        if_mode can be one of: 'direct', 'mixer', 'envelope_DC', 'envelope_AC','OFF'
        If only one value is given the setting is applied to both IF channels
        (I and Q) for the downconverter associated to element
        (how will we know that? shouldn't this be per downconverter?)
        If if_mode is a tuple, then the IF stage will be assigned to each
        quadrature independently, i.e.:
        if_mode = ('direct', 'envelope_AC') will set the I channel to be
        direct and the Q channel to be 'envelope_AC'
        :param disable_warning:
        :param octave_input_port:
        :param if_mode_q:
        :param if_mode_i:
        """

        octave = self._get_client(octave_input_port)
        octave.rf_inputs[octave_input_port[1]].set_if_mode_i(if_mode_i.to_sdk())
        octave.rf_inputs[octave_input_port[1]].set_if_mode_q(if_mode_q.to_sdk())

    def calibrate(
        self,
        octave_output_port: Tuple[str, int],
        lo_if_frequencies_tuple_list: List[Tuple] = None,
        save_to_db=True,
        **kwargs,
    ) -> Dict[Tuple[int, int], CalibrationResult]:
        """
        calibrates IQ mixer associated with element

        :param octave_output_port:
        :param lo_if_frequencies_tuple_list: A list of LO/IF frequencies for which the
            calibration is to be performed [(LO1, IF1), (LO2, IF2), ...]
        :param save_to_db:
        """

        octave = self._get_client(octave_output_port)
        output_port = octave_output_port[1]
        calibration_input = 2

        sdk_lo_source = octave.rf_outputs[output_port].get_lo_source()
        lo_source = OctaveLOSource.from_sdk(sdk_lo_source)
        only_one_unique_lo = all(
            lo_if_frequencies_tuple_list[0][0] == tup[0]
            for tup in lo_if_frequencies_tuple_list
        )
        if lo_source != OctaveLOSource.Internal:
            if not only_one_unique_lo:
                raise Exception(
                    "If the LO source is external, "
                    "only one lo frequency is allowed in calibration"
                )

        # TODO fix loop bug and remove this error:
        if not only_one_unique_lo:
            raise ValueError("multiple lo frequencies are not yed supported")

        optimizer_parameters = self._get_optimizer_parameters(
            kwargs.get("optimizer_parameters", None),
        )
        first_lo, first_if = lo_if_frequencies_tuple_list[0]
        compiled, qm = self._compile_calibration_program(
            first_lo, first_if, octave_output_port, optimizer_parameters
        )

        state_name_before = self._set_octave_for_calibration(
            octave, calibration_input, output_port
        )

        result = {}
        current_lo = first_lo
        for lo_freq, if_freq in lo_if_frequencies_tuple_list:
            if current_lo != lo_freq and lo_source == OctaveLOSource.Internal:
                octave.rf_outputs[output_port].set_lo_frequency(
                    lo_source.to_sdk(), lo_freq
                )  # if lo_source is not internal then don't do anything
                current_lo = lo_freq

            octave.rf_inputs[calibration_input].set_lo_source(
                RFInputLOSource.Analyzer.to_sdk()
            )

            octave.rf_inputs[calibration_input].set_lo_frequency(
                RFInputLOSource.Analyzer.to_sdk(),
                lo_freq + optimizer_parameters["calibration_offset_frequency"],
            )

            self._set_if_freq(qm, if_freq, optimizer_parameters)

            dc_offsets, correction = _run_compiled(compiled, qm)
            temp = octave.rf_outputs[output_port].get_temperature()

            result[lo_freq, if_freq] = CalibrationResult(
                mixer_id=octave_output_mixer_name(*octave_output_port),
                correction=correction,
                i_offset=dc_offsets[0],
                q_offset=dc_offsets[1],
                temperature=temp,
                if_frequency=if_freq,
                lo_frequency=lo_freq,
                optimizer_parameters=optimizer_parameters,
            )

            if save_to_db and self._octave_config.calibration_db is not None:
                self._octave_config.calibration_db.update_calibration_data(
                    result[lo_freq, if_freq]
                )

        # set to previous state
        octave.restore_state(state_name_before)

        return result

    def _get_optimizer_parameters(self, optimizer_parameters):
        optimizer_parameters_defaults = {
            "average_iterations": 100,
            "iterations": 10000,
            "calibration_offset_frequency": 7e6,
            "keep_on": False,
        }
        if optimizer_parameters is not None:
            for p in optimizer_parameters:
                if p in optimizer_parameters_defaults:
                    optimizer_parameters_defaults[p] = optimizer_parameters[p]
                else:
                    raise ValueError(f"optimizer parameter {p} is not supported")
        return optimizer_parameters_defaults

    def _compile_calibration_program(
        self,
        first_lo,
        first_if,
        octave_output_port,
        optimizer_parameters_defaults,
    ):
        from qm.octave._calibration_program import _generate_program

        iq_channels = self._octave_config.get_opx_iq_ports(octave_output_port)
        controller_name = iq_channels[0][0]
        adc_channels = [(controller_name, 1), (controller_name, 2)]
        config = _prep_config(
            iq_channels,
            adc_channels,
            first_if,
            first_lo,
            optimizer_parameters_defaults,
        )
        prog = _generate_program(optimizer_parameters_defaults)
        qm = self._qmm.open_qm(config)
        compiled = qm.compile(prog)

        return compiled, qm

    def _set_if_freq(self, qm, if_freq, optimizer_parameters):
        down_mixer_offset, signal_freq, image_freq = _get_frequencies(
            if_freq, optimizer_parameters
        )
        qm.set_intermediate_frequency("IQmixer", if_freq)
        qm.set_intermediate_frequency("signal_analyzer", signal_freq)
        qm.set_intermediate_frequency("lo_analyzer", down_mixer_offset)
        qm.set_intermediate_frequency("image_analyzer", image_freq)

    def _set_octave_for_calibration(
        self, octave, calibration_input, output_port
    ) -> str:
        state_name_before = "before_cal"
        octave.snapshot_state(state_name_before)
        # switch to loopback mode to listen in on the RF output
        octave.rf_inputs[calibration_input].set_rf_source(
            _convert_rf_output_index_to_input_source(output_port).to_sdk()
        )
        octave.rf_inputs[calibration_input].set_if_mode_i(IFMode.direct.to_sdk())
        octave.rf_inputs[calibration_input].set_if_mode_q(IFMode.direct.to_sdk())
        octave.rf_inputs[1].set_if_mode_i(IFMode.off.to_sdk())
        octave.rf_inputs[1].set_if_mode_q(IFMode.off.to_sdk())
        octave.rf_outputs[output_port].set_output(RFOutputMode.on.to_sdk())
        return state_name_before



from typing import Dict, Tuple, Union, List, Optional, Any

from qm.octave.calibration_db import CalibrationResult
from qm.octave.enums import (
    OctaveLOSource,
    RFOutputMode,
    RFInputLOSource,
    IFMode,
    ClockType,
    ClockFrequency,
    ClockInfo,
)
from qm.octave.octave_manager import OctaveManager


class QmOctave:
    def __init__(self, qm, octave_manager) -> None:
        super().__init__()
        self._qm = qm
        self._octave_manager: OctaveManager = octave_manager
        self._elements_to_rf_in: Dict[str, Tuple[str, int]] = {}
        self._elements_lo_frequency: Dict[str, float] = {}
        self._elements_lo_sources: Dict[str, OctaveLOSource] = {}

    def _get_element_opx_output(
        self, element
    ) -> Tuple[Tuple[str, int], Tuple[str, int]]:
        opx_config = self._qm.get_config()
        if element not in opx_config["elements"]:
            raise ValueError(f"could not find element {element}")
        if "mixInputs" not in opx_config["elements"][element]:
            raise ValueError(f"Element {element} is not of type mixInputs")
        i_port = opx_config["elements"][element]["mixInputs"]["I"]
        q_port = opx_config["elements"][element]["mixInputs"]["Q"]
        return i_port, q_port

    def _get_element_lo_frequency(
        self, element
    ) -> Tuple[Tuple[str, int], Tuple[str, int]]:
        opx_config = self._qm.get_config()
        if element not in opx_config["elements"]:
            raise ValueError(f"could not find element {element}")
        if "mixInputs" not in opx_config["elements"][element]:
            raise ValueError(f"Element {element} is not of type mixInputs")
        frequency = opx_config["elements"][element]["mixInputs"]["lo_frequency"]
        return frequency

    def _get_element_octave_output_port(self, element: str) -> (int, Any):
        opx_i_port, opx_q_port = self._get_element_opx_output(element)
        return self._octave_manager.get_output_port(opx_i_port, opx_q_port)

    def _get_lo_from_element_config(self, element):
        opx_config = self._qm.get_config()
        if element not in opx_config["elements"]:
            raise ValueError(f"could not find element {element}")
        if "mixInputs" not in opx_config["elements"][element]:
            raise ValueError(f"Element {element} is not of type mixInputs")
        lo_frequency = opx_config["elements"][element]["mixInputs"]["lo_frequency"]
        return lo_frequency

    def _get_if_from_element_config(self, element):
        opx_config = self._qm.get_config()
        if element not in opx_config["elements"]:
            raise ValueError(f"could not find element {element}")
        if_frequency = opx_config["elements"][element]["intermediate_frequency"]
        return if_frequency

    def _get_element_downconversion_input_port_and_client(
        self, element: str
    ) -> (str, int):
        if element in self._elements_to_rf_in:
            name, port = self._elements_to_rf_in[element]
            return name, port
        else:
            raise ValueError("Element has no readout port associated")

    def set_qua_element_octave_rf_in_port(self, element, octave_name, rf_input_index):
        """
        Sets the octave downconversion port for the element

        :param element:
        :param octave_name:
        :param rf_input_index:
        :return:
        """
        opx_config = self._qm.get_config()
        if element not in opx_config["elements"]:
            raise ValueError(f"could not find element {element}")
        self._elements_to_rf_in[element] = (octave_name, rf_input_index)

    def load_lo_frequency_from_config(self, elements: Union[List, str]):
        """
        Loads into the octave synthesizers the LO frequencies specified for elements
        in the element list

        :param elements:
        """
        opx_config = self._qm.get_config()
        for element in elements:
            if element not in opx_config["elements"]:
                raise ValueError(f"could not find element {element}")
            if "mixInputs" not in opx_config["elements"][element]:
                raise ValueError(f"Element {element} is not of type mixInputs")

        for element in elements:
            lo_frequency = opx_config["elements"][element]["mixInputs"]["lo_frequency"]
            self.set_lo_frequency(element, lo_frequency)

    def set_lo_frequency(
        self, element: str, lo_frequency: float, set_source: bool = True
    ):
        """
        Sets the LO frequency of the synthesizer associated to element

        :param element:
        :param lo_frequency:
        :param set_source:
        """
        octave_port = self._get_element_octave_output_port(element)
        lo_source = self._elements_lo_sources.get(element, OctaveLOSource.Internal)
        self._octave_manager.set_lo_frequency(
            octave_output_port=octave_port,
            lo_frequency=lo_frequency,
            lo_source=lo_source,
            set_source=set_source,
        )

    def update_external_lo_frequency(
        self,
        element: str,
        lo_frequency: float,
    ):
        """
        Updates Octave on the external LO frequency (provided by the user)
        associated with element

        :param element:
        :param lo_frequency:
        """
        self._elements_lo_frequency[element] = lo_frequency

    def set_lo_source(self, element: str, lo_port: OctaveLOSource):
        """
        Sets the source of LO going to the upconverter associated with element.

        :param element:
        :param lo_port:
        """
        octave_port = self._get_element_octave_output_port(element)
        self._elements_lo_sources[element] = lo_port
        self._octave_manager.set_lo_source(octave_port, lo_port)

    def set_rf_output_mode(self, element: str, switch_mode: RFOutputMode):
        """
        Configures the output switch of the upconverter associated to element.
        switch_mode can be either: 'always_on', 'always_off', 'normal' or 'inverted'
        When in 'normal' mode a high trigger will turn the switch on and a low trigger will turn it off
        When in 'inverted' mode a high trigger will turn the switch off and a low trigger will turn it on
        When in 'always_on' the switch will be permanently on. When in 'always_off' mode the switch will be permanently off.

        :param element:
        :param switch_mode:
        """
        octave_port = self._get_element_octave_output_port(element)
        self._octave_manager.set_rf_output_mode(octave_port, switch_mode)

    def set_rf_output_gain(self, element: str, gain_in_db: float):
        """
        Sets the RF output gain for the upconverter associated with element.
        RF_gain is in steps of 0.5dB from -X to +Y (X and Y TBD) and is referring
        to the maximum OPX output power of 4dBm (=0.5V pk-pk) So for a value of -24
        for example, an IF signal coming from the OPX at max power (4dBm) will be
        upconverted and come out of Octave at -20dBm

        :param element:
        :param gain_in_db:
        """
        # a. use value custom set in qm.octave.update_external
        # b. use value from config
        if element in self._elements_lo_frequency:
            frequency = self._elements_lo_frequency[element]
        else:
            frequency = self._get_element_lo_frequency(element)

        octave_port = self._get_element_octave_output_port(element)
        self._octave_manager.set_rf_output_gain(octave_port, gain_in_db, frequency)

    def set_downconversion(
        self,
        element: str,
        lo_source: Optional[RFInputLOSource] = None,
        lo_frequency: Optional[float] = None,
        if_mode_i: IFMode = IFMode.direct,
        if_mode_q: IFMode = IFMode.direct,
        disable_warning: bool = False,
    ):
        """
        Sets the LO source for the downconverters.
        The LO source will be the one associated with the upconversion of element

        :param element:
        :param lo_source:
        :param lo_frequency:
        :param if_mode_i:
        :param if_mode_q:
        :param disable_warning:
        """
        self._set_downconversion_lo(
            element, lo_source, lo_frequency, disable_warning=disable_warning
        )
        self._set_downconversion_if_mode(
            element, if_mode_i, if_mode_q, disable_warning=disable_warning
        )

    def _set_downconversion_lo(
        self,
        element: str,
        lo_source: Optional[RFInputLOSource] = None,
        lo_frequency: Optional[float] = None,
        disable_warning=False,
    ):
        """
        Sets the LO source for the downconverters.
        If no value is given the LO source will be the one associated with the
        upconversion of element

        :param element:
        :param lo_frequency:
        :param lo_source:
        :param disable_warning:
        """
        # TODO check if the downconversion shared between elements(from same rf_in or both rf_ins)
        # if we have two elements for the same downconversion:
        # warning when setting for port that has multiple elements
        # “element x shares the input port with elements [y],
        # any settings applied will affect all elements”
        octave_input_port = self._get_element_downconversion_input_port_and_client(
            element
        )
        if lo_source is None:
            octave_port = self._get_element_octave_output_port(element)
            rf_output_port = octave_port[1]
            if rf_output_port == 1:
                lo_source = RFInputLOSource.RFOutput1_LO
            elif rf_output_port == 2:
                lo_source = RFInputLOSource.RFOutput2_LO
            elif rf_output_port == 3:
                lo_source = RFInputLOSource.RFOutput3_LO
            elif rf_output_port == 4:
                lo_source = RFInputLOSource.RFOutput4_LO
            elif rf_output_port == 5:
                lo_source = RFInputLOSource.RFOutput5_LO
            else:
                raise ValueError(f"RF output {rf_output_port} is not found")

        self._octave_manager.set_downconversion_lo_source(
            octave_input_port, lo_source, lo_frequency, disable_warning
        )

    def _set_downconversion_if_mode(
        self,
        element: str,
        if_mode_i: IFMode = IFMode.direct,
        if_mode_q: IFMode = IFMode.direct,
        disable_warning=False,
    ):
        octave_input_port = self._get_element_downconversion_input_port_and_client(
            element
        )
        self._octave_manager.set_downconversion_if_mode(
            octave_input_port, if_mode_i, if_mode_q, disable_warning
        )

    def calibrate_element(
        self,
        element: str,
        lo_if_frequencies_tuple_list: Optional[List[Tuple]] = None,
        save_to_db=True,
        **kwargs,
    ) -> Dict[Tuple[int, int], CalibrationResult]:
        """
        Gets an element with a list of LO and IF couples to calibrate the mixers for

        :param element:
        :param lo_if_frequencies_tuple_list:
        :param save_to_db:
        :param kwargs:
        :return:
        """

        octave_port = self._get_element_octave_output_port(element)

        # get IF and LO frequencies from element
        if lo_if_frequencies_tuple_list is None:
            if_frequency = self._get_if_from_element_config(element)
            lo_frequency = self._get_lo_from_element_config(element)
            lo_if_frequencies_tuple_list = [(if_frequency, lo_frequency)]

        return self._octave_manager.calibrate(
            octave_port, lo_if_frequencies_tuple_list, save_to_db, **kwargs
        )

    def set_clock(
        self,
        octave_name: str,
        clock_type: ClockType,
        frequency: Optional[ClockFrequency],
    ):
        """
        This function will set the octave clock type - internal, external or buffered.
        It can also set the clock frequency - 10, 100 or 1000 MHz

        :param octave_name: str
        :param clock_type: ClockType
        :param frequency: ClockFrequency
        """
        # TODO: get name from the QMOctave instance
        self._octave_manager.set_clock(octave_name, clock_type, frequency)

    def get_clock(self, octave_name: str) -> ClockInfo:
        return self._octave_manager.get_clock(octave_name)
